@@ -1,8 +1,44 @@
+//!
+//! This engine is intended to provide behavior over time
+//! for one or more RGB LEDs (though could also be useful
+//! for any kind of color sequencing).
+//!
+//! In most cases:
+//!
+//! * Each LED will get a [`Sequence`] of some max length `N`.
+//! * Each [`Sequence`] contains up to `N` [`Action`]s
+//! * Each [`Action`] consists of:
+//!   * A [`Context`], which contains information like the color,
+//!       start time, and duration of an Action
+//!   * A [behavior], which is what the LED will do over time
+//!   * A [`LoopBehavior`], which describes whether the individual
+//!       Action will repeat in some way
+//! * The [`Sequence`] also contains a [`LoopBehavior`], which describes
+//!       whether the ENTIRE SEQUENCE will repeat in some way
+//!
+//! A typical usage example would be to:
+//!
+//! 1. Create an array of [`Sequence`]s, one for each LED
+//! 1. Use either the [`ActionBuilder`] or [`script!()`] macro
+//!     to define each [`Action`] in each [`Sequence`]
+//! 1. Periodically poll each [`Sequence`], using the resulting
+//!     color to update the physical LED
+//! 1. Modify or replace the [`Sequence`]s as necessary, for example
+//!     at a regular interval, or in response to some event, such as
+//!     a button press or received packet
+//!
+//! [`Sequence`]: crate::engine::Sequence
+//! [`Action`]: crate::engine::Action
+//! [`Context`]: crate::engine::Context
+//! [behavior]: crate::behaviors
+//! [`script!()`]: crate::script
+//! [`ActionBuilder`]: crate::engine::ActionBuilder
+//! [`LoopBehavior`]: crate::engine::LoopBehavior
+
 use core::cmp::min;
 use core::marker::PhantomData;
 use core::ops::{Deref, DerefMut};
 
-use crate::behaviors::AutoIncr;
 use crate::behaviors::{Cycler, FadeColor, SeekColor, StayColor};
 use crate::LossyIntoF32;
 use groundhog::RollingTimer;
@@ -10,26 +46,41 @@ use heapless::Vec;
 use smart_leds::colors::BLACK;
 use smart_leds::RGB8;
 
+/// A sequence of [`Action`]s with maximum length N.
+///
+/// The total [`Sequence`] also has a [`LoopBehavior`]
+/// that describes the looping behavior of the entire
+/// sequence (e.g. play the sequence N times, once,
+/// forever).
+///
+/// [`Action`]: Action
+/// [`LoopBehavior`]: LoopBehavior)
+/// [`Sequence`]: Sequence
 #[derive(Clone)]
 pub struct Sequence<R, const N: usize> {
     seq: Vec<Action<R>, N>,
     position: usize,
-    behavior: Behavior,
+    behavior: LoopBehavior,
     never_run: bool,
 }
 
 impl<R, const N: usize> Sequence<R, N> {
     const INIT: Sequence<R, N> = Sequence::new();
 
+    /// Create a new, empty sequence
     pub const fn new() -> Self {
         Self {
             seq: Vec::new(),
             position: 0,
-            behavior: Behavior::Nop,
+            behavior: LoopBehavior::Nop,
             never_run: true,
         }
     }
 
+    /// Create an array of new, empty sequences.
+    ///
+    /// This is often useful when creating an array for multiple
+    /// LEDs
     pub const fn new_array<const M: usize>() -> [Self; M] {
         [Self::INIT; M]
     }
@@ -44,124 +95,33 @@ where
     }
 }
 
-#[derive(Clone)]
-pub struct Action<R> {
-    action: Actions<R>,
-    behavior: Behavior,
-}
-
-#[derive(Clone, Default)]
-pub struct Context<R> {
-    pub(crate) start_tick: u32, // TODO: Hack - Not R::Tick because const init
-    pub(crate) auto_incr_phase: AutoIncr,
-    pub(crate) period_ms: f32,
-    pub(crate) duration_ms: u32, // TODO: Hack - Not R::Tick because const init
-    pub(crate) phase_offset_ms: u32, // TODO: Hack - Not R::Tick because const init
-    pub(crate) last_color: RGB8,
-    pub(crate) color: RGB8,
-    _pd: PhantomData<R>,
-}
-
-impl<R> Context<R>
-where
-    R: RollingTimer<Tick = u32> + Default + Clone,
-{
-    pub fn calc_end(&self) -> R::Tick {
-        self.start_tick
-            .wrapping_add(self.duration_ms * (R::TICKS_PER_SECOND / 1000))
-    }
-
-    pub fn calc_end_phase(&self) -> R::Tick {
-        self.phase_offset_ms.wrapping_add(self.duration_ms)
-    }
-
-    pub fn reinit(&mut self, start: R::Tick, start_ph: R::Tick, last_color: RGB8) {
-        self.start_tick = start;
-        self.last_color = last_color;
-        match self.auto_incr_phase {
-            AutoIncr::Never => {}
-            AutoIncr::Once => {
-                self.phase_offset_ms = start_ph;
-                self.auto_incr_phase = AutoIncr::Never;
-            }
-            AutoIncr::Forever => {
-                self.phase_offset_ms = start_ph;
-            }
-        }
-    }
-}
-
-#[derive(Clone)]
-pub struct Actions<R> {
-    context: Context<R>,
-    kind: ActionsKind,
-}
-
-#[derive(Clone)]
-pub enum ActionsKind {
-    Sin(Cycler),
-    Static(StayColor),
-    Fade(FadeColor),
-    Seek(SeekColor),
-}
-
-#[derive(Clone)]
-pub enum Behavior {
-    OneShot,
-    LoopForever,
-
-    #[allow(dead_code)]
-    LoopN {
-        current: usize,
-        cycles: usize,
-    },
-    Nop,
-}
-
-pub struct ActionBuilder<R>
-where
-    R: RollingTimer<Tick = u32> + Default + Clone,
-{
-    act: Action<R>,
-}
-
-impl Default for Behavior {
-    fn default() -> Self {
-        Behavior::Nop
-    }
-}
-
-impl<R> Default for Actions<R>
-where
-    R: RollingTimer<Tick = u32> + Default + Clone,
-{
-    fn default() -> Self {
-        Self {
-            context: Context::default(),
-            kind: ActionsKind::Static(StayColor::new()),
-        }
-    }
-}
-
 impl<R, const N: usize> Sequence<R, N>
 where
     R: RollingTimer<Tick = u32> + Default + Clone,
 {
+    /// Create a new, empty sequence
     pub fn empty() -> Self {
         Self {
             seq: Vec::new(),
             position: 0,
-            behavior: Behavior::OneShot,
+            behavior: LoopBehavior::OneShot,
             never_run: true,
         }
     }
 
+    /// Clear the current set of actions, leaving the
+    /// Sequence empty
     pub fn clear(&mut self) {
         self.seq.clear();
         self.position = 0;
     }
 
-    pub fn set(&mut self, actions: &[Action<R>], behavior: Behavior) {
+    /// Clear the current set of actions, and fill the sequence with
+    /// the given new actions.
+    ///
+    /// If `actions` is larger than the capacity of `self`, remaining items
+    /// will be ignored
+    pub fn set(&mut self, actions: &[Action<R>], behavior: LoopBehavior) {
         let amt = min(N, actions.len());
         self.clear();
 
@@ -171,6 +131,10 @@ where
         self.behavior = behavior;
     }
 
+    /// Poll the currently active Action, potentially also moving
+    /// to the next Action if necessary.
+    ///
+    /// When any Action is active, an RGB8 will be returned
     pub fn poll(&mut self) -> Option<RGB8> {
         if self.seq.is_empty() || (self.position >= self.seq.len()) {
             return None;
@@ -189,7 +153,7 @@ where
             self.never_run = false;
         }
 
-        use Behavior::*;
+        use LoopBehavior::*;
         match behavior {
             OneShot => seq[*position].poll().or_else(|| {
                 let end = seq[*position].calc_end();
@@ -244,6 +208,37 @@ where
     }
 }
 
+/// A single behavior step
+///
+/// An Action is a single describable step in a sequence
+/// of events.
+///
+/// * Each [`Action`] consists of:
+///   * A [`Context`], which contains information like the color,
+///       start time, and duration of an Action
+///   * A [behavior], which is what the LED will do over time
+///   * A [`LoopBehavior`], which describes whether the individual
+///       Action will repeat in some way
+///
+/// Actions are typically constructed through either the
+/// [`ActionBuilder`] or [`script!()`] macro.
+///
+/// Actions are not usually interacted with directly, but rather
+/// as part of a [`Sequence`].
+///
+/// [`Sequence`]: crate::engine::Sequence
+/// [`Action`]: crate::engine::Action
+/// [`Context`]: crate::engine::Context
+/// [behavior]: crate::behaviors
+/// [`script!()`]: crate::script
+/// [`ActionBuilder`]: crate::engine::ActionBuilder
+/// [`LoopBehavior`]: crate::engine::LoopBehavior
+#[derive(Clone)]
+pub struct Action<R> {
+    action: InnerAction<R>,
+    behavior: LoopBehavior,
+}
+
 impl<R> Deref for Action<R>
 where
     R: RollingTimer<Tick = u32> + Default + Clone,
@@ -259,18 +254,16 @@ impl<R> Action<R>
 where
     R: RollingTimer<Tick = u32> + Default + Clone,
 {
-    pub fn new(action: Actions<R>, behavior: Behavior) -> Self {
-        Self { action, behavior }
-    }
-
+    /// Return an ActionBuilder structure to configure a new
+    /// Action
     pub fn build() -> ActionBuilder<R> {
         ActionBuilder::new()
     }
 
-    pub fn reinit(&mut self, start: R::Tick, end_ph: R::Tick, last_color: RGB8) {
+    pub(crate) fn reinit(&mut self, start: R::Tick, end_ph: R::Tick, last_color: RGB8) {
         self.action.reinit(start, end_ph, last_color);
 
-        use Behavior::*;
+        use LoopBehavior::*;
         match &mut self.behavior {
             OneShot => {}
             LoopForever => {}
@@ -283,8 +276,8 @@ where
         }
     }
 
-    pub fn poll(&mut self) -> Option<RGB8> {
-        use Behavior::*;
+    pub(crate) fn poll(&mut self) -> Option<RGB8> {
+        use LoopBehavior::*;
 
         let action = &mut self.action;
         let behavior = &mut self.behavior;
@@ -315,16 +308,155 @@ where
     }
 }
 
-pub enum PhaseIncr {
-    Millis(u32),
-    AutoIncr,
-    AutoIncrOnStart,
+/// The behavior-independent information of an [`Action`]
+///
+/// The Context structure contains information like the color,
+/// start time, and duration of an [`Action`].
+///
+/// It is not usually necessary to interact with a Context directly.
+///
+/// [`Action`]: crate::engine::Action
+#[derive(Clone, Default)]
+pub struct Context<R> {
+    pub(crate) start_tick: u32, // TODO: Hack - Not R::Tick because const init
+    pub(crate) auto_incr_phase: AutoIncr,
+    pub(crate) period_ms: f32,
+    pub(crate) duration_ms: u32, // TODO: Hack - Not R::Tick because const init
+    pub(crate) phase_offset_ms: u32, // TODO: Hack - Not R::Tick because const init
+    pub(crate) last_color: RGB8,
+    pub(crate) color: RGB8,
+    _pd: PhantomData<R>,
 }
 
-impl From<u32> for PhaseIncr {
-    fn from(data: u32) -> Self {
-        PhaseIncr::Millis(data)
+impl<R> Context<R>
+where
+    R: RollingTimer<Tick = u32> + Default + Clone,
+{
+    pub(crate) fn calc_end(&self) -> R::Tick {
+        self.start_tick
+            .wrapping_add(self.duration_ms * (R::TICKS_PER_SECOND / 1000))
     }
+
+    pub(crate) fn calc_end_phase(&self) -> R::Tick {
+        self.phase_offset_ms.wrapping_add(self.duration_ms)
+    }
+
+    pub(crate) fn reinit(&mut self, start: R::Tick, start_ph: R::Tick, last_color: RGB8) {
+        self.start_tick = start;
+        self.last_color = last_color;
+        match self.auto_incr_phase {
+            AutoIncr::Never => {}
+            AutoIncr::Once => {
+                self.phase_offset_ms = start_ph;
+                self.auto_incr_phase = AutoIncr::Never;
+            }
+            AutoIncr::Forever => {
+                self.phase_offset_ms = start_ph;
+            }
+        }
+    }
+}
+
+#[derive(Clone)]
+struct InnerAction<R> {
+    context: Context<R>,
+    kind: InnerActionKind,
+}
+
+impl<R> Default for InnerAction<R>
+where
+    R: RollingTimer<Tick = u32> + Default + Clone,
+{
+    fn default() -> Self {
+        Self {
+            context: Context::default(),
+            kind: InnerActionKind::Static(StayColor::new()),
+        }
+    }
+}
+
+impl<R> Deref for InnerAction<R>
+where
+    R: RollingTimer<Tick = u32> + Default + Clone,
+{
+    type Target = Context<R>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.context
+    }
+}
+
+impl<R> DerefMut for InnerAction<R>
+where
+    R: RollingTimer<Tick = u32> + Default + Clone,
+{
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.context
+    }
+}
+
+impl<R> InnerAction<R>
+where
+    R: RollingTimer<Tick = u32> + Default + Clone,
+{
+    pub fn poll(&self) -> Option<RGB8> {
+        use InnerActionKind::*;
+        match &self.kind {
+            Sin(s) => s.poll(&self.context),
+            Static(s) => s.poll(&self.context),
+            Fade(f) => f.poll(&self.context),
+            Seek(s) => s.poll(&self.context),
+        }
+    }
+}
+
+#[derive(Clone)]
+enum InnerActionKind {
+    Sin(Cycler),
+    Static(StayColor),
+    Fade(FadeColor),
+    Seek(SeekColor),
+}
+
+/// A description of the looping behavior of an [`Action`] or [`Sequence`]
+///
+/// [`Sequence`]: crate::engine::Sequence
+/// [`Action`]: crate::engine::Action
+#[derive(Clone)]
+pub enum LoopBehavior {
+    /// Execute this action or sequence exactly once
+    OneShot,
+
+    /// Loop this action or sequence endlessly
+    LoopForever,
+
+    /// Loop this action or sequence N times
+    LoopN {
+        /// The current iteration
+        current: usize,
+
+        /// The total number of iterations
+        cycles: usize,
+    },
+
+    /// This action will immediately yield to the next
+    Nop,
+}
+
+impl Default for LoopBehavior {
+    fn default() -> Self {
+        LoopBehavior::Nop
+    }
+}
+
+/// A builder for the [`Action`] structure
+///
+/// [`Action`]: crate::engine::Action
+pub struct ActionBuilder<R>
+where
+    R: RollingTimer<Tick = u32> + Default + Clone,
+{
+    act: Action<R>,
 }
 
 // Builder Methods
@@ -332,48 +464,55 @@ impl<R> ActionBuilder<R>
 where
     R: RollingTimer<Tick = u32> + Default + Clone,
 {
+    /// Create a new ActionBuilder with default settings
     #[inline(always)]
     pub fn new() -> Self {
         Self {
             act: Action {
-                action: Actions::default(),
-                behavior: Behavior::default(),
+                action: InnerAction::default(),
+                behavior: LoopBehavior::default(),
             },
         }
     }
 
+    /// Finalize the ActionBuilder into an Action
     #[inline(always)]
     pub fn finish(self) -> Action<R> {
         self.act
     }
 
+    /// Set the LoopBehavior to repeat `ct` times
     #[inline(always)]
     pub fn times(mut self, ct: usize) -> Self {
-        self.act.behavior = Behavior::LoopN {
+        self.act.behavior = LoopBehavior::LoopN {
             current: 0,
             cycles: ct,
         };
         self
     }
 
+    /// Set the LoopBehavior to repeat never
     #[inline(always)]
     pub fn once(mut self) -> Self {
-        self.act.behavior = Behavior::OneShot;
+        self.act.behavior = LoopBehavior::OneShot;
         self
     }
 
+    /// Set the LoopBehavior to loop forever
     #[inline(always)]
     pub fn forever(mut self) -> Self {
-        self.act.behavior = Behavior::LoopForever;
+        self.act.behavior = LoopBehavior::LoopForever;
         self
     }
 
+    /// Set the color
     #[inline(always)]
     pub fn color(mut self, color: RGB8) -> Self {
         self.act.action.context.color = color;
         self
     }
 
+    /// Set the duration in milliseconds
     #[inline(always)]
     pub fn for_ms(mut self, duration: R::Tick) -> Self {
         self.act.action.context.duration_ms = duration;
@@ -381,12 +520,13 @@ where
         // TODO: This might be better to remove later? Probably
         // conside how to handle these "hacks", or abstract over
         // the cycler type more reasonably
-        if let ActionsKind::Fade(_) = self.act.action.kind {
+        if let InnerActionKind::Fade(_) = self.act.action.kind {
             self.act.action.context.period_ms = duration.lossy_into() * 4.0;
         }
         self
     }
 
+    /// Set the phase offset behavior
     #[inline(always)]
     pub fn phase_offset_ms(mut self, phase_offset_ms: PhaseIncr) -> Self {
         let (phase_offset_ms, incr) = match phase_offset_ms {
@@ -399,101 +539,107 @@ where
         self
     }
 
+    /// Set the duration and period in one step
     #[inline(always)]
     pub fn dur_per_ms(mut self, duration: R::Tick, period_ms: f32) -> Self {
         self.act.action.context.duration_ms = duration;
 
         // TODO: fix hax?
         self.act.action.context.period_ms = match self.act.action.kind {
-            ActionsKind::Sin(_) => period_ms * 2.0,
-            ActionsKind::Static(_) => period_ms,
-            ActionsKind::Fade(_) => duration.lossy_into() * 4.0,
-            ActionsKind::Seek(_) => period_ms,
+            InnerActionKind::Sin(_) => period_ms * 2.0,
+            InnerActionKind::Static(_) => period_ms,
+            InnerActionKind::Fade(_) => duration.lossy_into() * 4.0,
+            InnerActionKind::Seek(_) => period_ms,
         };
 
         self
     }
 
+    /// Set the period, in milliseconds (as an f32)
     #[inline(always)]
     pub fn period_ms(mut self, duration: f32) -> Self {
         self.act.action.context.period_ms = duration;
         self
     }
 
+    /// Convert the current ActionBuilder to produce a Sine Cycler
     #[inline(always)]
     pub fn sin(mut self) -> Self {
         let mut sin = Cycler::new();
         sin.start_low();
-        self.act.action.kind = ActionsKind::Sin(sin);
+        self.act.action.kind = InnerActionKind::Sin(sin);
         self
     }
 
+    /// Convert the current ActionBuilder to produce a SeekColor behavior
     #[inline(always)]
     pub fn seek(mut self) -> Self {
-        self.act.action.kind = ActionsKind::Seek(SeekColor);
+        self.act.action.kind = InnerActionKind::Seek(SeekColor);
         self
     }
 
+    /// Convert the current ActionBuilder to produce a Cosine Cycler
     #[inline(always)]
     pub fn cos(mut self) -> Self {
         let mut cos = Cycler::new();
         cos.start_high();
-        self.act.action.kind = ActionsKind::Sin(cos);
+        self.act.action.kind = InnerActionKind::Sin(cos);
         self
     }
 
+    /// Convert the current ActionBuilder to produce a StayColor action
     #[inline(always)]
     pub fn solid(mut self) -> Self {
-        self.act.action.kind = ActionsKind::Static(StayColor::new());
+        self.act.action.kind = InnerActionKind::Static(StayColor::new());
         self
     }
 
+    /// Convert the current ActionBuilder to produce a Fade Up action
     #[inline(always)]
     pub fn fade_up(mut self) -> Self {
         self.act.action.kind =
-            ActionsKind::Fade(FadeColor::new_fade_up(&mut self.act.action.context));
+            InnerActionKind::Fade(FadeColor::new_fade_up(&mut self.act.action.context));
         self
     }
 
+    /// Convert the current ActionBuilder to produce a Fade Down action
     #[inline(always)]
     pub fn fade_down(mut self) -> Self {
         self.act.action.kind =
-            ActionsKind::Fade(FadeColor::new_fade_down(&mut self.act.action.context));
+            InnerActionKind::Fade(FadeColor::new_fade_down(&mut self.act.action.context));
         self
     }
 }
 
-impl<R> Deref for Actions<R>
-where
-    R: RollingTimer<Tick = u32> + Default + Clone,
-{
-    type Target = Context<R>;
+/// A description of Phase Increment Behavior
+pub enum PhaseIncr {
+    /// A specific phase increment, in milliseconds
+    Millis(u32),
 
-    fn deref(&self) -> &Self::Target {
-        &self.context
+    /// Increment the phase from the last elapsed phase
+    /// on every action transition
+    AutoIncr,
+
+    /// Increment the phase from the last elapsed phase
+    /// on the first action transition
+    AutoIncrOnStart,
+}
+
+impl From<u32> for PhaseIncr {
+    fn from(data: u32) -> Self {
+        PhaseIncr::Millis(data)
     }
 }
 
-impl<R> DerefMut for Actions<R>
-where
-    R: RollingTimer<Tick = u32> + Default + Clone,
-{
-    fn deref_mut(&mut self) -> &mut Self::Target {
-        &mut self.context
-    }
+#[derive(Debug, Clone, Copy)]
+pub(crate) enum AutoIncr {
+    Never,
+    Once,
+    Forever,
 }
 
-impl<R> Actions<R>
-where
-    R: RollingTimer<Tick = u32> + Default + Clone,
-{
-    pub fn poll(&self) -> Option<RGB8> {
-        use ActionsKind::*;
-        match &self.kind {
-            Sin(s) => s.poll(&self.context),
-            Static(s) => s.poll(&self.context),
-            Fade(f) => f.poll(&self.context),
-            Seek(s) => s.poll(&self.context),
-        }
+impl Default for AutoIncr {
+    fn default() -> Self {
+        AutoIncr::Never
     }
 }
